@@ -3,8 +3,9 @@
 
 from __future__ import absolute_import, division, print_function
 
-import io
+import asyncio
 from glob import has_magic
+import io
 import logging
 import os
 import warnings
@@ -15,6 +16,7 @@ from azure.storage.blob._shared.base_client import create_configuration
 from azure.datalake.store import AzureDLFileSystem, lib
 from azure.datalake.store.core import AzureDLFile, AzureDLPath
 from azure.storage.blob.aio import BlobServiceClient as AIOBlobServiceClient
+from azure.storage.blob import generate_blob_sas, BlobSasPermissions
 from azure.storage.blob.aio._list_blobs_helper import BlobPrefix
 from azure.storage.blob._models import BlobBlock, BlobProperties, BlobType
 from fsspec import AbstractFileSystem
@@ -26,8 +28,14 @@ from fsspec.asyn import (
 )
 from fsspec.spec import AbstractBufferedFile
 from fsspec.utils import infer_storage_options, tokenize
-from .utils import filter_blobs
+from .utils import (
+    filter_blobs,
+    get_blob_metadata,
+    close_service_client,
+    close_container_client,
+)
 
+from datetime import datetime, timedelta
 
 logger = logging.getLogger(__name__)
 
@@ -42,6 +50,7 @@ FORWARDED_BLOB_PROPERTIES = [
     "remaining_retention_days",
     "archive_status",
     "last_accessed_on",
+    "etag",
     "tags",
     "tag_count",
 ]
@@ -370,15 +379,8 @@ class AzureBlobFileSystem(AsyncFileSystem):
             for k in ["use_listings_cache", "listings_expiry_time", "max_paths"]
             if k in kwargs
         }  # pass on to fsspec superclass
-        super().__init__(asynchronous=asynchronous, **super_kwargs)
-        if account_name:
-            self.account_name = account_name
-        elif "AZURE_STORAGE_ACCOUNT_NAME" in os.environ.keys():
-            self.account_name = os.environ["AZURE_STORAGE_ACCOUNT_NAME"]
-        else:
-            raise ValueError(
-                "account_name must be declared explictly or set in AZURE_STORAGE_ACCOUNT_NAME environmental variable"
-            )
+        super().__init__(asynchronous=asynchronous, loop=loop, **super_kwargs)
+        self.account_name = account_name or os.getenv("AZURE_STORAGE_ACCOUNT_NAME")
         self.account_key = account_key or os.getenv("AZURE_STORAGE_ACCOUNT_KEY")
         self.connection_string = connection_string or os.getenv(
             "AZURE_STORAGE_CONNECTION_STRING"
@@ -391,7 +393,6 @@ class AzureBlobFileSystem(AsyncFileSystem):
         self.request_session = request_session
         self.socket_timeout = socket_timeout
         self.blocksize = blocksize
-        self.loop = loop or get_loop()
         self.default_fill_cache = default_fill_cache
         self.default_cache_type = default_cache_type
         if (
@@ -407,7 +408,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         else:
             self.sync_credential = None
         self.do_connect()
-        weakref.finalize(self, maybe_sync, self.service_client.close, self)
+        weakref.finalize(self, maybe_sync, close_service_client, self, self)
 
     @classmethod
     def _strip_protocol(cls, path: str):
@@ -474,24 +475,35 @@ class AzureBlobFileSystem(AsyncFileSystem):
         ValueError if none of the connection details are available
         """
         try:
-            self.account_url: str = f"https://{self.account_name}.blob.core.windows.net"
-            creds = [self.credential, self.account_key]
-            if any(creds):
-                self.service_client = [
-                    AIOBlobServiceClient(account_url=self.account_url, credential=cred)
-                    for cred in creds
-                    if cred is not None
-                ][0]
-            elif self.connection_string is not None:
+            if self.connection_string is not None:
                 self.service_client = AIOBlobServiceClient.from_connection_string(
                     conn_str=self.connection_string
                 )
-            elif self.sas_token is not None:
-                self.service_client = AIOBlobServiceClient(
-                    account_url=self.account_url + self.sas_token, credential=None
-                )
+            elif self.account_name:
+                self.account_url: str = f"https://{self.account_name}.blob.core.windows.net"
+                creds = [self.credential, self.account_key]
+                if any(creds):
+                    self.service_client = [
+                        AIOBlobServiceClient(
+                            account_url=self.account_url, credential=cred
+                        )
+                        for cred in creds
+                        if cred is not None
+                    ][0]
+                elif self.sas_token is not None:
+                    if not self.sas_token.startswith("?"):
+                        self.sas_token = f"?{self.sas_token}"
+                    self.service_client = AIOBlobServiceClient(
+                        account_url=self.account_url + self.sas_token, credential=None
+                    )
+                else:
+                    self.service_client = AIOBlobServiceClient(
+                        account_url=self.account_url
+                    )
             else:
-                self.service_client = AIOBlobServiceClient(account_url=self.account_url)
+                raise ValueError(
+                    "Must provide either a connection_string or account_name with credentials!!"
+                )
 
         except Exception as e:
             raise ValueError(f"unable to connect to account for {e}")
@@ -528,12 +540,15 @@ class AzureBlobFileSystem(AsyncFileSystem):
             return path.split(delimiter, 1)
 
     def info(self, path, refresh=False, **kwargs):
-        fetch_from_azure = (path and self._ls_from_cache(path) is None) or refresh
+        try:
+            fetch_from_azure = (path and self._ls_from_cache(path) is None) or refresh
+        except Exception:
+            fetch_from_azure = True
         if fetch_from_azure:
-            return maybe_sync(self._info, self, path)
+            return maybe_sync(self._info, self, path, refresh)
         return super().info(path)
 
-    async def _info(self, path, **kwargs):
+    async def _info(self, path, refresh=False, **kwargs):
         """Give details of entry at path
         Returns a single dictionary, with exactly the same information as ``ls``
         would with ``detail=True``.
@@ -546,12 +561,18 @@ class AzureBlobFileSystem(AsyncFileSystem):
         dict with keys: name (full path in the FS), size (in bytes), type (file,
         directory, or something else) and other FS-specific keys.
         """
+        if refresh:
+            invalidate_cache = True
+        else:
+            invalidate_cache = False
         path = self._strip_protocol(path)
-        out = await self._ls(self._parent(path), **kwargs)
-        out = [o for o in out if o["name"].rstrip("/") == path]
+        out = await self._ls(
+            self._parent(path), invalidate_cache=invalidate_cache, **kwargs
+        )
+        out = [o for o in out if (o["name"].rstrip("/") + "/") == path]
         if out:
             return out[0]
-        out = await self._ls(path, **kwargs)
+        out = await self._ls(path, invalidate_cache=invalidate_cache, **kwargs)
         path = path.rstrip("/")
         out1 = [o for o in out if o["name"].rstrip("/") == path]
         if len(out1) == 1:
@@ -559,7 +580,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
                 out1[0]["size"] = None
             return out1[0]
         elif len(out1) > 1 or out:
-            return {"name": path, "size": 0, "type": "directory"}
+            return {"name": path, "size": None, "type": "directory"}
         else:
             raise FileNotFoundError
 
@@ -698,8 +719,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
 
         if invalidate_cache:
             self.dircache.clear()
-
-        if (container in ["", ".", delimiter]) and (path in ["", delimiter]):
+        if (container in ["", ".", "*", delimiter]) and (path in ["", delimiter]):
             if path not in self.dircache or invalidate_cache or return_glob:
                 # This is the case where only the containers are being returned
                 logger.info(
@@ -715,13 +735,13 @@ class AzureBlobFileSystem(AsyncFileSystem):
             if target_path not in self.dircache or invalidate_cache or return_glob:
                 if container not in ["", delimiter]:
                     # This is the case where the container name is passed
-                    container_client = self.service_client.get_container_client(
+                    async with self.service_client.get_container_client(
                         container=container
-                    )
-                    path = path.strip("/")
-                    blobs = container_client.walk_blobs(
-                        include=["metadata"], name_starts_with=path
-                    )
+                    ) as cc:
+                        path = path.strip("/")
+                        blobs = cc.walk_blobs(
+                            include=["metadata"], name_starts_with=path
+                        )
 
                     # Check the depth that needs to be screened
                     depth = target_path.count("/")
@@ -748,6 +768,13 @@ class AzureBlobFileSystem(AsyncFileSystem):
                                                 == depth
                                             ):
                                                 outblobs.append(blob_)
+                                            elif (
+                                                blob_["name"].count("/") == depth
+                                                and blob_["size"] == 0
+                                            ):
+                                                outblobs.append(blob_)
+                                            else:
+                                                pass
                                         elif blob_["name"].count("/") == (depth):
                                             outblobs.append(blob_)
                                         else:
@@ -793,7 +820,6 @@ class AzureBlobFileSystem(AsyncFileSystem):
         List of dicts
             Returns details about the contents, such as name, size and type
         """
-
         output = []
         for content in contents:
             data = {
@@ -803,20 +829,35 @@ class AzureBlobFileSystem(AsyncFileSystem):
             }
             if content.has_key("container"):  # NOQA
                 fname = f"{content.container}{delimiter}{content.name}"
-                data.update({"name": fname})
+                fname = fname.rstrip(delimiter)
                 if content.has_key("size"):  # NOQA
+                    data.update({"name": fname})
                     data.update({"size": content.size})
-                else:
-                    data.update({"size": 0})
-                if data["size"] == 0 and data["name"].endswith(delimiter):
-                    data.update({"type": "directory"})
-                else:
                     data.update({"type": "file"})
+                else:
+                    data.update({"name": fname})
+                    data.update({"size": None})
+                    data.update({"type": "directory"})
             else:
-                fname = f"{content.name}{delimiter}"
+                fname = f"{content.name}"
                 data.update({"name": fname})
-                data.update({"size": 0})
+                data.update({"size": None})
                 data.update({"type": "directory"})
+            if "metadata" in data.keys():
+                if data["metadata"] is not None:
+                    if (
+                        "is_directory" in data["metadata"].keys()
+                        and data["metadata"]["is_directory"] == "true"
+                    ):
+                        data.update({"type": "directory"})
+                        data.update({"size": None})
+                    elif (
+                        "is_directory" in data["metadata"].keys()
+                        and data["metadata"]["is_directory"] == "false"
+                    ):
+                        data.update({"type": "file"})
+                    else:
+                        pass
             if return_glob:
                 data.update({"name": data["name"].rstrip("/")})
             output.append(data)
@@ -825,6 +866,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
                 # This handles the case where path is a file passed to ls
                 return output
             output = await filter_blobs(output, target_path)
+
         return output
 
     def find(self, path, withdirs=False, prefix="", **kwargs):
@@ -832,7 +874,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
             self._find, self, path=path, withdirs=withdirs, prefix=prefix, **kwargs
         )
 
-    async def _find(self, path, withdirs=False, prefix="", **kwargs):
+    async def _find(self, path, withdirs=False, prefix="", with_parent=False, **kwargs):
         """List all files below path.
         Like posix ``find`` command without conditions
         Parameters
@@ -849,11 +891,14 @@ class AzureBlobFileSystem(AsyncFileSystem):
         path = self._strip_protocol(path)
         parent_path = path.strip("/") + "/"
         target_path = f"{parent_path}{(prefix or '').lstrip('/')}"
-        # target_path = path.strip("/")
         container, path = self.split_path(target_path)
 
-        container_client = self.service_client.get_container_client(container=container)
-        blobs = container_client.list_blobs(include=["metadata"], name_starts_with=path)
+        async with self.service_client.get_container_client(
+            container=container
+        ) as container_client:
+            blobs = container_client.list_blobs(
+                include=["metadata"], name_starts_with=path
+            )
         files = {}
         dir_set = set()
         dirs = {}
@@ -862,7 +907,9 @@ class AzureBlobFileSystem(AsyncFileSystem):
             infos = await self._details([b async for b in blobs])
             for info in infos:
                 name = info["name"]
-                parent_dir = self._parent(name)
+                if name == target_path:
+                    continue
+                parent_dir = self._parent(name).rstrip("/") + "/"
                 if parent_dir not in dir_set and parent_dir != parent_path.strip("/"):
                     dir_set.add(parent_dir)
                     dirs[parent_dir] = {
@@ -870,11 +917,17 @@ class AzureBlobFileSystem(AsyncFileSystem):
                         "type": "directory",
                         "size": 0,
                     }
-                files[name] = info
+                if info["type"] == "directory":
+                    dirs[name] = info
+                if info["type"] == "file":
+                    files[name] = info
+
         except ResourceNotFoundError:
             # find doesn't raise but returns [] or {} instead
             pass
         if withdirs:
+            if not with_parent:
+                dirs.pop(target_path, None)
             files.update(dirs)
         names = sorted(files)
         if not detail:
@@ -986,7 +1039,8 @@ class AzureBlobFileSystem(AsyncFileSystem):
 
     async def _mkdir(self, path, create_parents=True, delimiter="/", **kwargs):
         """
-        Create directory entry at path
+        Mkdir is a no-op for creating anything except top-level containers.
+        This aligns to the Azure Blob Filesystem flat hierarchy
 
         Parameters
         ----------
@@ -1000,42 +1054,42 @@ class AzureBlobFileSystem(AsyncFileSystem):
             Delimiter to use when splitting the path
 
         """
+        fullpath = path
         container_name, path = self.split_path(path, delimiter=delimiter)
         _containers = await self._ls("")
         _containers = [c["name"] for c in _containers]
-        # The list of containers will be returned from _ls() in a directory format,
-        # with a trailing "/", but the container_name will not have this.
+        # The list of containers will be returned from _ls() in a directory format
         # Need a placeholder that presents the container_name in a directory format
-        container_name_as_dir = f"{container_name}/"
         if _containers is None:
             _containers = []
         try:
-            if container_name_as_dir not in _containers:
+            if container_name not in _containers:
                 if create_parents:
                     # create new container
-                    await self.service_client.create_container(name=container_name)
-                    self.invalidate_cache(self._parent(path))
-                    if path:
-                        fpath = f"{container_name_as_dir}{path}"
-                        await self._mkdir(fpath, exist_ok=True)
+                    await self.service_client.create_container(
+                        name=container_name, metadata={"is_directory": "true"}
+                    )
+                    self.invalidate_cache(self._parent(container_name))
                 else:
                     raise PermissionError(
                         "Azure Container does not exist.  Set create_parents=True to create!!"
                     )
             else:
-                ## attempt to create prefix
-                container_client = self.service_client.get_container_client(
-                    container=container_name
-                )
-                await container_client.upload_blob(name=path, data="", overwrite=False)
+                exist_ok = kwargs.get("exist_ok", True)
+                if exist_ok:
+                    pass
+                else:
+                    raise
+
         except PermissionError:
             raise PermissionError(
-                f"Unable to create Azure container {container_name_as_dir} with \
+                f"Unable to create Azure container {container_name} with \
                 create_parents=False"
             )
+
         except Exception as e:
-            ## everything else
-            exist_ok = kwargs.get("exist_ok", False)
+            # everything else
+            exist_ok = kwargs.get("exist_ok", True)
             if exist_ok:
                 pass
             else:
@@ -1043,7 +1097,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
                     f"Cannot overwrite existing Azure container -- {container_name} already exists. \
                         with Azure error {e}"
                 )
-        self.invalidate_cache(self._parent(path))
+        self.invalidate_cache(self._parent(fullpath))
 
     mkdir = sync_wrapper(_mkdir)
 
@@ -1063,7 +1117,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
             If False (default), raise an error if the directory already exists.
         """
         try:
-            self.mkdir(path, create_parents=True)
+            self.mkdir(path, create_parents=True, exist_ok=exist_ok)
         except FileExistsError:
             if exist_ok:
                 pass
@@ -1084,14 +1138,16 @@ class AzureBlobFileSystem(AsyncFileSystem):
             If None, there will be no limit and infinite recursion may be
             possible.
         """
-        path = await self._expand_path(path, recursive=recursive, maxdepth=maxdepth)
+        path = await self._expand_path(
+            path, recursive=recursive, maxdepth=maxdepth, with_parent=True
+        )
         for p in reversed(path):
-            await self.rm_file(p)
+            await self._rm_file(p)
         self.invalidate_cache()
 
     rm = sync_wrapper(_rm)
 
-    async def rm_file(self, path, delimiter="/", **kwargs):
+    async def _rm_file(self, path, delimiter="/", **kwargs):
         """
         Delete a given file
 
@@ -1105,30 +1161,27 @@ class AzureBlobFileSystem(AsyncFileSystem):
         """
         try:
             kind = await self._info(path)
+            container_name, path = self.split_path(path, delimiter=delimiter)
             kind = kind["type"]
-            if kind == "file":
-                container_name, path = self.split_path(path, delimiter=delimiter)
-                container_client = self.service_client.get_container_client(
+            if path != "":
+                async with self.service_client.get_container_client(
                     container=container_name
-                )
-                logger.debug(f"Delete blob {path} in {container_name}")
-                await container_client.delete_blob(path)
+                ) as cc:
+                    await cc.delete_blob(path)
             elif kind == "directory":
-                container_name, path = self.split_path(path, delimiter=delimiter)
-                container_client = self.service_client.get_container_client(
-                    container=container_name
-                )
-                _containers = await self._ls("")
-                _containers = [c["name"] for c in _containers]
-                if (container_name + delimiter in _containers) and (not path):
-                    logger.debug(f"Delete container {container_name}")
-                    await container_client.delete_container()
+                await self._rmdir(container_name)
             else:
-                raise RuntimeError(f"Unable to delete {path}!")
-            self.invalidate_cache(self._parent(path))
-
+                raise RuntimeError(f"Unable to remove {path}")
+        except ResourceNotFoundError:
+            pass
         except FileNotFoundError:
             pass
+        except Exception as e:
+            raise RuntimeError(f"Failed to remove {path} for {e}")
+
+        self.invalidate_cache(self._parent(path))
+
+    sync_wrapper(_rm_file)
 
     def rmdir(self, path: str, delimiter="/", **kwargs):
         maybe_sync(self._rmdir, self, path, delimiter=delimiter, **kwargs)
@@ -1150,7 +1203,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         container_name, path = self.split_path(path, delimiter=delimiter)
         _containers = await self._ls("")
         _containers = [c["name"] for c in _containers]
-        if (container_name + delimiter in _containers) and (not path):
+        if (container_name in _containers) and (not path):
             # delete container
             await self.service_client.delete_container(container_name)
             self.invalidate_cache(self._parent(path))
@@ -1212,6 +1265,8 @@ class AzureBlobFileSystem(AsyncFileSystem):
                 return True
         except FileNotFoundError:
             pass
+        except KeyError:
+            pass
 
         container_name, path = self.split_path(path)
 
@@ -1219,11 +1274,8 @@ class AzureBlobFileSystem(AsyncFileSystem):
             # Empty paths exist by definition
             return True
 
-        cc = self.service_client.get_container_client(container_name)
-        bc = cc.get_blob_client(blob=path)
-        async with bc:
+        async with self.service_client.get_blob_client(container_name, path) as bc:
             exists = await bc.exists()
-
         return exists
 
     async def _pipe_file(self, path, value, overwrite=True, **kwargs):
@@ -1232,7 +1284,9 @@ class AzureBlobFileSystem(AsyncFileSystem):
         async with self.service_client.get_blob_client(
             container=container_name, blob=path
         ) as bc:
-            result = await bc.upload_blob(data=value, overwrite=overwrite)
+            result = await bc.upload_blob(
+                data=value, overwrite=overwrite, metadata={"is_directory": "false"}
+            )
         self.invalidate_cache(self._parent(path))
         return result
 
@@ -1268,13 +1322,53 @@ class AzureBlobFileSystem(AsyncFileSystem):
         else:
             return self.cat_file(paths[0])
 
+    def url(self, path, expires=3600, **kwargs):
+        return maybe_sync(self._url, self, path, expires, **kwargs)
+
+    async def _url(self, path, expires=3600, **kwargs):
+        """Generate presigned URL to access path by HTTP
+
+        Parameters
+        ----------
+        path : string
+            the key path we are interested in
+        expires : int
+            the number of seconds this signature will be good for.
+        """
+        container_name, blob = self.split_path(path)
+
+        sas_token = generate_blob_sas(
+            account_name=self.account_name,
+            container_name=container_name,
+            blob_name=blob,
+            account_key=self.account_key,
+            permission=BlobSasPermissions(read=True),
+            expiry=datetime.utcnow() + timedelta(seconds=expires),
+        )
+
+        async with self.service_client.get_blob_client(container_name, blob) as bc:
+            url = f"{bc.url}?{sas_token}"
+        return url
+
     def expand_path(self, path, recursive=False, maxdepth=None):
         return maybe_sync(self._expand_path, self, path, recursive, maxdepth)
 
-    async def _expand_path(self, path, recursive=False, maxdepth=None):
+    async def _expand_path(self, path, recursive=False, maxdepth=None, **kwargs):
         """Turn one or more globs or directories into a list of all matching files"""
+
+        with_parent = kwargs.get(
+            "with_parent", False
+        )  # Sets whether to return the parent dir
+
+        if isinstance(path, list):
+            path = [f"{p.strip('/')}/" for p in path if not p.endswith("*")]
+        else:
+            if not path.endswith("*"):
+                path = f"{path.strip('/')}/"
         if isinstance(path, str):
-            out = await self._expand_path([path], recursive, maxdepth)
+            out = await self._expand_path(
+                [path], recursive, maxdepth, with_parent=with_parent
+            )
         else:
             out = set()
             path = [self._strip_protocol(p) for p in path]
@@ -1287,11 +1381,23 @@ class AzureBlobFileSystem(AsyncFileSystem):
                         out |= bit2
                     continue
                 elif recursive:
-                    rec = set(await self._find(p, withdirs=True))
+                    rec = set(
+                        await self._find(p, withdirs=True, with_parent=with_parent)
+                    )
                     out |= rec
-                if p not in out and (recursive is False or await self._exists(p)):
-                    # only check once for root
+
+                if p not in out and (
+                    recursive is False
+                    or await self._exists(p)
+                    or await self._exists(p.rstrip("/"))
+                ):
+                    if not await self._exists(p):
+                        # This is to verify that we don't miss files
+                        p = p.rstrip("/")
+                        if not await self._exists(p):
+                            continue
                     out.add(p)
+
         if not out:
             raise FileNotFoundError
         return list(sorted(out))
@@ -1307,17 +1413,21 @@ class AzureBlobFileSystem(AsyncFileSystem):
         """
 
         container_name, path = self.split_path(rpath, delimiter=delimiter)
-        cc = self.service_client.get_container_client(container_name)
-        bc = cc.get_blob_client(blob=path)
+
         if os.path.isdir(lpath):
             self.makedirs(rpath, exist_ok=True)
         else:
             try:
                 with open(lpath, "rb") as f1:
-                    await bc.upload_blob(f1, overwrite=overwrite)
+                    async with self.service_client.get_blob_client(
+                        container_name, path
+                    ) as bc:
+                        await bc.upload_blob(
+                            f1, overwrite=overwrite, metadata={"is_directory": "false"}
+                        )
                 self.invalidate_cache()
             except ResourceExistsError:
-                raise FileExistsError("File already exists!!")
+                raise FileExistsError("File already exists!")
             except ResourceNotFoundError:
                 if not await self._exists(container_name):
                     raise FileNotFoundError("Container does not exist.")
@@ -1358,16 +1468,16 @@ class AzureBlobFileSystem(AsyncFileSystem):
         files = [f["name"] for f in files]
         container_name, path = self.split_path(rpath, delimiter=delimiter)
         try:
-            cc = self.service_client.get_container_client(container_name)
-            bc = cc.get_blob_client(blob=path)
-
             if await self._isdir(rpath):
                 os.makedirs(lpath, exist_ok=True)
             else:
-                with open(lpath, "wb") as my_blob:
-                    stream = await bc.download_blob()
-                    data = await stream.readall()
-                    my_blob.write(data)
+                async with self.service_client.get_blob_client(
+                    container_name, path
+                ) as bc:
+                    with open(lpath, "wb") as my_blob:
+                        stream = await bc.download_blob()
+                        data = await stream.readall()
+                        my_blob.write(data)
         except Exception as e:
             raise FileNotFoundError(f"File not found for {e}")
 
@@ -1380,9 +1490,8 @@ class AzureBlobFileSystem(AsyncFileSystem):
     async def _setxattrs(self, rpath, **kwargs):
         container_name, path = self.split_path(rpath)
         try:
-            cc = self.service_client.get_container_client(container_name)
-            bc = cc.get_blob_client(blob=path)
-            await bc.set_blob_metadata(metadata=kwargs)
+            async with self.service_client.get_blob_client(container_name, path) as bc:
+                await bc.set_blob_metadata(metadata=kwargs)
             self.invalidate_cache(self._parent(rpath))
         except Exception as e:
             raise FileNotFoundError(f"File not found for {e}")
@@ -1538,8 +1647,12 @@ class AzureBlobFile(AbstractBufferedFile):
             self.cache = caches[cache_type](
                 self.blocksize, self._fetch_range, self.size, **cache_options
             )
+            self.metadata = maybe_sync(
+                get_blob_metadata, self, self.container_client, self.blob
+            )
+
         else:
-            self.metadata = metadata
+            self.metadata = metadata or {"is_directory": "false"}
             self.buffer = io.BytesIO()
             self.offset = None
             self.forced = False
@@ -1547,7 +1660,7 @@ class AzureBlobFile(AbstractBufferedFile):
 
     def close(self):
         """Close file and azure client."""
-        maybe_sync(self.container_client.close, self)
+        asyncio.run_coroutine_threadsafe(close_container_client(self), loop=self.loop)
         super().close()
 
     def connect_client(self):
@@ -1587,7 +1700,7 @@ class AzureBlobFile(AbstractBufferedFile):
         except Exception as e:
             raise ValueError(
                 f"Unable to fetch container_client with provided params for {e}!!"
-            )
+            ) from e
 
     async def _async_fetch_range(self, start: int, end: int, **kwargs):
         """
@@ -1700,3 +1813,10 @@ class AzureBlobFile(AbstractBufferedFile):
             )
 
     _upload_chunk = sync_wrapper(_async_upload_chunk)
+
+    def __del__(self):
+        try:
+            if not self.closed:
+                self.close()
+        except TypeError:
+            pass
